import * as exception from "./exception";
import * as env from "./environment";
import * as pino from "pino";
export interface ILoggerOptions {
    /** if not specified, defaults to env.logLevel */
    level?: env.LogLevel;
    /** ignore some keys when printing to console.   defaults to "pid,hostname"
     * disabled for UAT or PROD
    */
    ignore?: string;
    /** defaults to "yyyymmdd_HHMM:ss.l"
     * if set to false, will be unix epoch tick.
     * disabled for UAT or PROD
     * @remarks
     * [format docs here](https://www.npmjs.com/package/dateformat)
    */
    translateTime?: string | boolean;
    /**
     * **browser only**:  under what envLevel browser calls will redirect to console (not to the logging subsstem).
     * This is useful for development debugging, as the browser has clickable source links
     * **IMPORTANT**: using this skips,  the internal logging subsystem: just outputs to console.
     * default is ["dev"]
     * used in conjunction with {@link browserConsoleLogLevel} (both must be true)
     * @remarks
     *
     */
    browserConsoleEnvLevel?: env.EnvLevel[];
    /**
     * **browser only**:  under what logLevel browser calls will redirect to console (not to the logging subsstem).
     * This is useful for development debugging, as the browser has clickable source links.
     * **IMPORTANT**: using this skips,  the internal logging subsystem: just outputs to console.
     * default is ["trace","debug"]
     * used in conjunction with {@link browserConsoleEnvLevel} (both must be true)
     * @remarks
     *
     */
    browserConsoleLogLevel?: env.LogLevel[];
}
export declare class Logger {
    name: string;
    options: ILoggerOptions;
    /**
     * logging subsystem.   see https://github.com/pinojs/pino
     */
    _pino: pino.Logger;
    constructor(name: string, options?: ILoggerOptions);
    /** used internally but exposed for convenience.
     * allows calling pino as the logger does.
     * If you want to access pino directly, use the _pino member
     */
    private _callPino;
    trace(msgOrObj: string | ILogObj): void;
    trace(obj: ILogObj, msg: string): void;
    debug(msgOrObj: string | ILogObj): void;
    debug(obj: ILogObj, msg: string): void;
    info(msgOrObj: string | ILogObj): void;
    info(obj: ILogObj, msg: string): void;
    warn(msgOrObj: string | ILogObj): void;
    warn(obj: ILogObj, msg: string): void;
    error(msgOrObj: string | ILogObj): void;
    error(obj: ILogObj, msg: string): void;
    fatal(msgOrObj: string | ILogObj): void;
    fatal(obj: ILogObj, msg: string): void;
    silent(msgOrObj: string | ILogObj): void;
    silent(obj: ILogObj, msg: string): void;
    assert(condition: boolean, msgOrObj: string | ILogObj): void;
    assert(condition: boolean, obj: ILogObj, msg: string): void;
}
interface ILogObj {
    [key: string]: unknown;
    /** if provided, will be considered the "main" message for presentation purposes */
    msg?: string;
    /** if provided, logs error details
     * any other error objects will be truncated from the log.
     * @remarks
     * If you want to include other error objects you need to manually serialize them via ```xlib.diagnostics.exception.errorToJson()```
    */
    err?: Error;
}
/** @internal helper to throw a XlibException if the condition !==true */
export declare function throwCheck(condition: boolean, errorMsg?: string): void;
/** thrown on race-check failures */
export declare class RaceCheckException extends exception.XlibException {
}
/** DEBUG ONLY!  noops self if environment.logLevel is greater than DEBUG helper to check for algorithm race conditions (example: removing from a collection while enumerating) */
export declare class DebugRaceCheck {
    debugText: string;
    private _version;
    private _lockVersion;
    private _isDisabled;
    constructor(debugText?: string);
    enter(): void;
    exit(): void;
    /** quickly locks and unlocks */
    edit(): void;
    /** ensures that this raceCheck is not currently locked */
    poke(): void;
    toString(): string;
}
export {};
