import * as types from "./_internal/types";
/** shape of all errors.   either derived from the ```Error``` object, or Error objects serialized to JSON */
export interface IError {
    /** the name of the Error class (typeName) */
    name: string;
    /** human readable and ***actionable*** error message */
    message: string;
    /** while almost always available, it may not be set under unusual circumstances */
    stack?: string;
    /** may be set if the error was created via xlib.diagnostics.Exception  */
    innerError?: IError;
    /** may be set if the error was created via xlib.diagnostics.Exception  */
    details?: {
        [key: string]: unknown;
    };
}
/** the shape of Errors that xlib serializes (the same as normal Error serialization, except the stack is an array, not a single string)*/
export interface IErrorJson {
    /** the name of the Error class (typeName) */
    name: string;
    /** human readable and ***actionable*** error message */
    message: string;
    /** while almost always available, it may not be set under unusual circumstances */
    stack?: string[];
    /** optional, can pass an innerException of you use xlib.diagnostics.Exception */
    innerError?: IErrorJson;
    /** may be set if the error was created via xlib.diagnostics.Exception  */
    details?: {
        [key: string]: unknown;
    };
}
/** passed to {@link Exception} constructor */
export interface IExceptionOptions {
    innerError?: Error;
    /** truncate logging frames from the top of the stack.
     * a good way to remove logging/util functions from the trace. */
    stackFramesToTruncate?: number;
    /** if you wish to restrict the number of stack frames stored, set this.   by default all stack frames are stored. */
    maxStackFrames?: number;
    /** additional details you want present in the error object.  found under .details */
    details?: {
        [key: string]: unknown;
    };
}
export declare function getStackTrace(): string[];
/** An `Error` object that includes improved features
 * @remarks
 * ## important usage
 * If you derive from this and have a custom constructor, be sure to call the base class's `super()` in your constructor.  This will ensure that your derived class inherits from `Error`
 *
 * @example
 * ```typescript
 * class MyException extends Exception{ } throw new MyException("boo");
 * ```
 *
 * @see
core idea from https://www.typescriptlang.org/docs/handbook/2/classes.html#inheriting-built-in-types

*/ export declare class Exception extends Error {
    message: string;
    stack: string;
    private static _getTypeNameOrFuncNameRegex;
    innerError?: Error;
    details?: {
        [key: string]: unknown;
    };
    constructor(message: string, options?: IExceptionOptions);
    /** includes stack track in string*/
    toString(): string;
    toJson<T extends Error = this>(): types.PropsUnion<IErrorJson, Pick<T, { [K in keyof T]: T[K] extends Function ? never : K; }[keyof T]>>;
}
/** all errors thrown by xlib extend this error type */
export declare class XlibException extends Exception {
}
/**
 * an exception that includes a statusCode for returning via http requests
 */
export declare class HttpStatusCodeException extends Exception {
    statusCode: number;
    constructor(message: string, statusCode: number, options: IExceptionOptions);
}
/** ensures you have an Error object.
    * If you pass in an error, you will get it back.  If you pass in anything else, a new Error will be created and a string represtation of your input will be set as the ```message``` parameter.
 @remarks
    * useful for ```try{}catch(ex){}``` statements, where you are uncertain what the thrown error is.
    * @example
 ```typescript
    * try{
    *  await someClass.someMethod();
    * }catch(_err){
    *  const err = __.diag.toError(_err);
    * //use err as a normal Error object
    * }
 ```
*/
export declare function toError(ex: ANY | Error): Error;
/** get a string representation of the error */
export declare function errorToString(ex: Error | IError, options?: IErrorToJsonOptions): string;
export interface IErrorToJsonOptions {
    /** max stack depth to output
        * @default Infinity
     */
    maxStacks?: number;
    /** by default, we will hide the stack in UAT or PROD envLevel, unless DEBUG logLevel is set.    pass TRUE to never hide the stack.
        * @default false
    */
    alwaysShowFullStack?: boolean;
    /** by default, we will hide the extra properties if the exception ctor options specify it.  pass TRUE to never hide properties */
    alwaysShowProperties?: boolean;
}
declare type ErrorAsJson<TError extends Error> = types.PropsUnion<IErrorJson, types.PropsRemove<TError, Function>>;
/** convert an error and all it's properties to JSON.   */
export declare function errorToJson<TError extends Error>(error: TError | IError, options?: IErrorToJsonOptions): ErrorAsJson<TError>;
export {};
//# sourceMappingURL=exception.d.ts.map