/**
 * @packageDocumentation
 * utilize http consistently across browser and node
 */
import { gaxios } from "./_imports";
export { gaxios };
import * as promise from "./promise";
export declare type IRemoteHttpEndpointOptions<TRecievePayload = ANY> = IRemoteHttpEndpointOverrideOptions<TRecievePayload> & {
    /** by default, all HTTP requests are made as soon as they are requested.  pass autoscaler options if your endpoint supports autoscaling.
    *
    IMPORTANT NOTE:  The remote endpoint Needs to return either HTTP 429 or 503 errors to represent ```"BACKOFF"``` messages.  all other errors will be treated normally.*/
    autoscalerOptions?: promise.IAutoscalerOptions;
};
export interface IRemoteHttpEndpointOverrideOptions<TRecievePayload = ANY> {
    /** the endpoint URL you will connect to */
    endpoint?: {
        origin?: string;
        path?: string;
    };
    /** if you want to retry failed requests.   from: https://www.npmjs.com/package/bluebird-retry
        * Options are:
    interval initial wait time between attempts in milliseconds (default 1000)

backoff if specified, increase interval by this factor between attempts

max_interval if specified, maximum amount that interval can increase to

timeout total time to wait for the operation to succeed in milliseconds

max_tries maximum number of attempts to try the operation (default 5)

predicate to be used as bluebird's Filtered Catch. func will be retried only if the predicate expectation is met, it will otherwise fail immediately.

throw_original to throw the last thrown error instance rather then a timeout error.
        * default is no retries. */
    retryOptions?: promise.IRetryOptions;
    /** optional settings sent with the request.  from: https://www.npmjs.com/package/gaxios
    */
    requestOptions?: gaxios.GaxiosOptions;
    /** inspect the response from the server and decide what to do:
            return  ```ABORT``` to ABORT RETRY (stop immediately with the error passed to reject())
            return ```RETRY``` to retry the request according to retryOptions
            return ```OK``` to stop and succeed with the current server response.
            DEFAULT: by default, if you don't specify your own ```responseInspector``` function, will insert a function that will retry error 500 and above (assuming ```retryOptions``` are set) */
    responseInspector?: (
    /** note: network issues are converted into err.response so you don't need to parse them differently.*/
    result: IGaxiosResult) => PromiseLike<"RETRY" | "ABORT" | "OK">;
}
export interface IGaxiosResult {
    err?: gaxios.GaxiosError;
    response?: gaxios.GaxiosResponse;
}
/**
*  define a remote http endpoint for reuse in your application
* includes retry logic and exponential backoff.
*/
export declare class RemoteHttpEndpoint<TSubmitPayload, TRecievePayload> {
    /** default options that will be applied to all requests.  may be overridden by each actual .post() or .get() call */
    defaultOptions: IRemoteHttpEndpointOptions<TRecievePayload>;
    /** autoscaler created if the constructor is passed autoscaler options */
    private _autoscaler?;
    private _gaxios;
    constructor(
    /** default options that will be applied to all requests.  may be overridden by each actual .post() or .get() call */
    defaultOptions?: IRemoteHttpEndpointOptions<TRecievePayload>);
    _onTooBusy: Array<(endpoint: RemoteHttpEndpoint<TSubmitPayload, TRecievePayload>) => any>;
    toJson(): {
        options: IRemoteHttpEndpointOptions<TRecievePayload>;
        autoscaler: {
            pendingCalls: number;
            activeCalls: number;
            metrics: {
                maxActive: number;
                tooBusyWaitStart: Date;
                activeCount: number;
                lastGrow: Date;
                lastMax: Date;
                lastTooBusy: Date;
                lastDecay: Date;
            };
            options: promise.IAutoscalerOptions;
        } | undefined;
    };
    post(
    /** pass a payload to POST */
    submitPayload: TSubmitPayload, 
    /**override any default options if desired */
    overrideOptions?: IRemoteHttpEndpointOverrideOptions): Promise<gaxios.GaxiosResponse<TRecievePayload>>;
    get(
    /** pass a payload to POST */
    /**override any default options if desired */
    overrideOptions?: IRemoteHttpEndpointOverrideOptions): Promise<gaxios.GaxiosResponse<TRecievePayload>>;
    private _doRequest;
    private _doRequest_send;
}
//# sourceMappingURL=net.d.ts.map