import * as exception from "./exception";
import * as rx from "rxjs";
export { rx };
/** a promise that resolves after an amount of time.   good for delaying execution
 */
export declare function delay(ms: Numeric): Promise<void>;
/** returns a new promise that will reject with a {@link TimeoutRejectError} if the initial promise does not resolve in the required timeframe
 * Important:  if a timeout occurs, the initial promise is not aborted.  It may later resolve successfully even though it won't be listened to.
 */
export declare function timeout<TResult>(ms: Numeric, initial: PromiseLike<TResult>, rejectMessage?: string): Promise<TResult>;
/** rejection raised by {@ timeout} */
export declare class TimeoutRejectError extends exception.XlibException {
}
/** options passes to the retry method
 * @see retry  */
export interface IRetryOptions {
    /**  return true to accept the result (or error).  if not provided, will return false when a promise failure occurs (retry will be attempted) */
    acceptResult?: <TResult, TPromise extends PromiseLike<TResult>>(details: {
        isSuccess: boolean;
        err?: Error;
        result?: TResult;
        promise: TPromise;
    }) => boolean;
    maxRetries: number;
    minDelay?: number;
    maxDelay?: number;
    /** default 100 */
    maxJitter?: number;
    expGrowDelay?: number;
    /** if errors occur and all retries are unsuccessful, we default by throwing a {@link RetryExceededError}.  set to true to throw first error returned instead. */
    throw_original?: boolean;
}
/** error that is thrown by the [[retry]] method if number of retries is exceeded.  You may also manually reject with this to abort execution of the retry() logic */
export declare class RetryStopError<TResult> extends exception.XlibException {
}
/**
 * retries a failed promise untill success or exit condition occurs.
 * to abort early, throw or reject with an { @link xlib.promise.RetryStopError }
 * @param options
 * @param fn
 */
export declare function retry<TResult>(options: IRetryOptions, fn: () => PromiseLike<TResult>): Promise<TResult>;
export interface IInspectablePromise<TResult> extends Promise<TResult> {
    status(): IPromiseStatus<TResult>;
}
interface IPromiseStatus<TResult> {
    /** true is the promise is not fulfilled or rejected */
    isPending: boolean;
    /** true = promise fullfilled successfully.
     * false = promise rejected
     * undefined = promise is still pending
     */
    isSuccess?: boolean;
    result?: TResult;
    error?: Error;
}
/**
 * @deprecated
 * use {@link exposeStatus} instead
 * @obsolete
 * */
export declare const awaitInspect: typeof exposeStatus;
/** adds a ```.status()``` property to your existing promise, allowing inspection of it's current state.   If the promise already has a ```status``` property, an XlibException is thrown. */
export declare function exposeStatus<TResult>(promise: PromiseLike<TResult>): IInspectablePromise<TResult>;
export interface IMutablePromise<TResult = void> extends IInspectablePromise<TResult> {
    resolve: (result: TResult) => void;
    reject: (error: Error) => void;
}
/** create a promise who's resolve/reject occur from an outside system.
 * @example
 * //for example, resolve the promise by calling
 * let myProm = createMutablePromise();
 * myProm.resolve();
 *  */
export declare function createMutablePromise<TResult = void>(): (Promise<TResult> & IMutablePromise<TResult>);
/** @deprecated use [[createMutablePromise]] instead */
export declare const CreateExposedPromise: typeof createMutablePromise;
/**  @deprecated use [[IMutablePromise]] instead */
export declare type IExposedPromise = IMutablePromise;
/** @deprecated use [[ReadWriteLock]] instead  */
export declare type AsyncReaderWriterLock = ReadWriteLock;
/** Synchronization construct.  Allows for exclusive writes and parallel reads.  requests are processed in a FIFO order.
 *
 * Usage notes:
 * - You should hold a lock for as little time as possible to avoid blocking other users.
 * - Be sure to call .[[readEnd]] or .[[writeEnd]] to avoid deadlocks
 * @example
 * const lock = new ReadWriteLock("start")
 * try{
 * 	const initialValue = lock.writeBegin()
 * 	console.log("initialValue==='start'",initialValue==="start")
 * }finally{
 * 	lock.writeEnd("finish")
 * }
 *
 */
export declare class ReadWriteLock<TValue = void> {
    /** multiple reads can occur at the same time */
    private _currentReads;
    /** requests are processed FIFO.  pending requests are stored here. */
    private _pendingTransactions;
    /** if the lock sychronizes data, it can be stored here */
    private _value;
    private _currentWrite?;
    /**
     *
     * @param value optional data to synchronize access to
     */
    constructor(value?: TValue);
    private _tryProcessNext;
    read<TReturn>(readFcn: (value: TValue) => void | PromiseLike<void>): Promise<void>;
    write(writeFcn: (value: TValue) => (TValue | PromiseLike<TValue>)): Promise<void>;
    tryReadBegin(): ({
        lockAquired: false;
    } | {
        lockAquired: true;
        value: TValue;
    });
    readBegin(): Promise<TValue>;
    readEnd(): void;
    tryWriteBegin(): ({
        lockAquired: false;
    } | {
        lockAquired: true;
        value: TValue;
    });
    writeBegin(): Promise<TValue>;
    writeEnd(value: TValue | PromiseLike<TValue>): Promise<void>;
}
export interface IAutoscalerOptions {
    /** minimum parallel requests (maxActive) you allow, regardless of how long the autoscaler has been idle.  should be 1 or more.
    */
    minParallel: number;
    /** optional.  set a max to number of parallel requests (maxActive) no matter how active the calls
        * @default undefined (no limit)
    */
    maxParallel?: number;
    /** if we get a "TOO_BUSY" rejection (from the ```failureListener```), how long we should wait before trying to expand our maxActive again. */
    busyGrowDelayMs: number;
    /** when we are at max parallel and still able to successfully submit requests (not getting "TOO_BUSY" errors), how long to delay before increasing our maxActive by 1. */
    growDelayMs: number;
    /** when we are under our max parallel, how long before our max should decrease by 1 .   Also, when we are consistently getting "TOO_BUSY" rejections, we will decrease our maxActive by 1 this often.  pass null to never decay (not recomended).*/
    idleOrBusyDecreaseMs: number;
    /** optional.  when we first get a "TOO_BUSY" rejection, we will reduce maxActive by this amount.  interval to check if we should penalize resets after ```busyWaitMs```
         * Note: when too busy, we also reduce maxActive via the ```decayDelayMs``` parameter every so often (as set by decayDelayMs)..   set to 0 to have no penalty except that set by decayDelayMs
        * @default 1
     */
    busyExtraPenalty?: number;
}
export declare class Autoscaler<TResult, TWorkerFunc extends (...args: ANY[]) => PromiseLike<TResult>> {
    private options;
    private backendWorker;
    /** will be used to intercept failures (promise rejections) from the ```backendWorker``` function.  should return "FAIL" if it's a normal failure (to be returned to the caller) or "TOO_BUSY" if the request should be retried  */
    private failureListener;
    private _metrics;
    private pendingCalls;
    private activeCalls;
    /** submit a request to the backend worker.
     *
     * **Important note**: to avoid "unhandled promise rejections" you need to make sure the returned Promise has a catch() applied to it.
     * **NOT** just store the promise in an array to inspect later.  This is because if the request fails, the returned promise gets rejected, and if the Promise internal logic doesn't see a .catch() it will show the global "unhandled rejected promse" soft error message.
     */
    submitRequest: TWorkerFunc;
    private _lastTryCallTime;
    constructor(options: IAutoscalerOptions, backendWorker: TWorkerFunc, 
    /** will be used to intercept failures (promise rejections) from the ```backendWorker``` function.  should return "FAIL" if it's a normal failure (to be returned to the caller) or "TOO_BUSY" if the request should be retried  */
    failureListener: ((err: Error) => "FAIL" | "TOO_BUSY"));
    toJson(): {
        pendingCalls: number;
        activeCalls: number;
        metrics: {
            /** the max number of active parallel requests we currently allow.   increases and decreases based on the growDelayMs and decayDelayMs */
            maxActive: number;
            /** time in which we decided to stop growing (based on options.busyWaitMs ) */
            tooBusyWaitStart: Date;
            /** the current number of parallel requests active in our backendWorker */
            activeCount: number;
            /** the last time we grew our maxActive count  */
            lastGrow: Date;
            /** the last time we were at our maxActive count */
            lastMax: Date;
            /** the last time we got a "TOO_BUSY" rejection from the backendWorker.  note that this could happen while in a options.busyWaitMs interval, if the backend is sufficently overwhelmed */
            lastTooBusy: Date;
            /** the last time we decayed our maxActive */
            lastDecay: Date;
        };
        options: IAutoscalerOptions;
    };
    private _tryCallBackend;
    private _heartbeatHandle;
}
//# sourceMappingURL=promise.d.ts.map